source(g_currentModDirectory .. "BaleWrapperStateEvent.lua")
source(g_currentModDirectory .. "BaleFix.lua")
BaleWrapper = {}
BaleWrapper.STATE_NONE = 0
BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER = 1
BaleWrapper.STATE_MOVING_GRABBER_TO_WORK = 2
BaleWrapper.STATE_WRAPPER_WRAPPING_BALE = 3
BaleWrapper.STATE_WRAPPER_FINSIHED = 4
BaleWrapper.STATE_WRAPPER_DROPPING_BALE = 5
BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM = 6
BaleWrapper.STATE_NUM_BITS = 3
BaleWrapper.CHANGE_GRAB_BALE = 1
BaleWrapper.CHANGE_DROP_BALE_AT_GRABBER = 2
BaleWrapper.CHANGE_WRAPPING_START = 3
BaleWrapper.CHANGE_WRAPPING_BALE_FINSIHED = 4
BaleWrapper.CHANGE_WRAPPER_START_DROP_BALE = 5
BaleWrapper.CHANGE_WRAPPER_BALE_DROPPED = 6
BaleWrapper.CHANGE_WRAPPER_PLATFORM_RESET = 7
BaleWrapper.CHANGE_BUTTON_EMPTY = 8
function BaleWrapper.prerequisitesPresent(specializations)
  return SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations)
end
function BaleWrapper:load(xmlFile)
  self.getIsFoldAllowed = Utils.overwrittenFunction(self.getIsFoldAllowed, BaleWrapper.getIsFoldAllowed)
  self.allowsGrabbingBale = BaleWrapper.allowsGrabbingBale
  self.updateWrappingState = BaleWrapper.updateWrappingState
  self.dropBaleFromWrapper = BaleWrapper.dropBaleFromWrapper
  self.moveBaleToWrapper = BaleWrapper.moveBaleToWrapper
  self.doStateChange = BaleWrapper.doStateChange
  self.updateWrapNodes = BaleWrapper.updateWrapNodes
  local wrapper = {}
  wrapper.baleNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.wrapper#baleIndex"))
  wrapper.wrapperNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.wrapper#index"))
  local wrappingAnimCurve = AnimCurve:new(linearInterpolatorN)
  local keyI = 0
  for keyI=0,2 do
    local key = string.format("vehicle.wrapper.key(%d)", keyI)
    local t = getXMLFloat(xmlFile, key .. "#time")
    local baleX, baleY, baleZ = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#baleRot"))
    if baleX ~= nil and baleY ~= nil then
      if baleZ == nil then
        break
      end
      baleX = math.rad(Utils.getNoNil(baleX, 0))
      baleY = math.rad(Utils.getNoNil(baleY, 0))
      baleZ = math.rad(Utils.getNoNil(baleZ, 0))
      local wrapperX, wrapperY, wrapperZ = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#wrapperRot"))
      wrapperX = math.rad(Utils.getNoNil(wrapperX, 0))
      wrapperY = math.rad(Utils.getNoNil(wrapperY, 0))
      wrapperZ = math.rad(Utils.getNoNil(wrapperZ, 0))
      wrappingAnimCurve:addKeyframe({
        v = {
          baleX,
          baleY,
          baleZ,
          wrapperX,
          wrapperY,
          wrapperZ
        },
        time = t
      })
	end
  end
  wrapper.animCurve = wrappingAnimCurve
  wrapper.animTime = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.wrapper#wrappingTime"), 5) * 1000
  wrapper.currentTime = 0
  wrapper.currentBale = nil
  wrapper.foldMinLimit = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.wrapper#foldMinLimit"), 0)
  wrapper.foldMaxLimit = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.wrapper#foldMaxLimit"), 1)
  self.wrapper = wrapper
  self.showInvalidBaleWarning = false
  self.allowedBaleTypes = {}
  local i = 0
  while true do
    local key = string.format("vehicle.wrapper.baleType(%d)", i)
    if not hasXMLProperty(xmlFile, key) then
      break
    end
    local wrapperBaleFilename = Utils.getFilename(getXMLString(xmlFile, key .. "#wrapperBaleFilename"), self.baseDirectory)
    local fillTypeStr = getXMLString(xmlFile, key .. "#fillType")
    if fillTypeStr ~= nil then
      local fillType = Fillable.fillTypeNameToInt[fillTypeStr]
      local minBaleDiameter = getXMLFloat(xmlFile, key .. "#minBaleDiameter")
      local maxBaleDiameter = getXMLFloat(xmlFile, key .. "#maxBaleDiameter")
      local minBaleWidth = getXMLFloat(xmlFile, key .. "#minBaleWidth")
      local maxBaleWidth = getXMLFloat(xmlFile, key .. "#maxBaleWidth")
      if wrapperBaleFilename ~= nil and minBaleDiameter ~= nil and maxBaleDiameter ~= nil and minBaleWidth ~= nil and maxBaleWidth ~= nil and fillType ~= nil then
        if self.allowedBaleTypes[fillType] == nil then
          self.allowedBaleTypes[fillType] = {}
        end
        table.insert(self.allowedBaleTypes[fillType], {
          wrapperBaleFilename = wrapperBaleFilename,
          minBaleDiameter = minBaleDiameter,
          maxBaleDiameter = maxBaleDiameter,
          minBaleWidth = minBaleWidth,
          maxBaleWidth = maxBaleWidth,
          fillType = fillType
        })
      end
    end
    i = i + 1
  end
  local wrapAnimNodes = {}
  local wrapAnimNodeI = 0
  while true do
    local wrapAnimNodeKey = string.format("vehicle.wrapper.wrapAnimNode(%d)", wrapAnimNodeI)
    if not hasXMLProperty(xmlFile, wrapAnimNodeKey) then
      break
    end
    local nodeId = Utils.indexToObject(self.components, getXMLString(xmlFile, wrapAnimNodeKey .. "#index"))
    if nodeId ~= nil then
      local animCurve = AnimCurve:new(linearInterpolatorN)
      local keyI = 0
      while true do
        local key = string.format(wrapAnimNodeKey .. ".key(%d)", keyI)
        local wrapperRot = getXMLFloat(xmlFile, key .. "#wrapperRot")
        if wrapperRot == nil then
          break
        end
        local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#trans"))
        local rx, ry, rz = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#rot"))
        local sx, sy, sz = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#scale"))
        x = Utils.getNoNil(x, 0)
        y = Utils.getNoNil(y, 0)
        z = Utils.getNoNil(z, 0)
        rx = math.rad(Utils.getNoNil(rx, 0))
        ry = math.rad(Utils.getNoNil(ry, 0))
        rz = math.rad(Utils.getNoNil(rz, 0))
        sx = Utils.getNoNil(sx, 1)
        sy = Utils.getNoNil(sy, 1)
        sz = Utils.getNoNil(sz, 1)
        animCurve:addKeyframe({
        v = {
          x,
          y,
          z,
          rx,
          ry,
          rz,
          sx,
          sy,
          sz
        },
        time = math.rad(wrapperRot)
      })
      keyI = keyI + 1
    end
    if keyI > 0 then
      local repeatWrapperRot = Utils.getNoNil(getXMLBool(xmlFile, wrapAnimNodeKey .. "#repeatWrapperRot"), false)
      table.insert(wrapAnimNodes, {
        nodeId = nodeId,
        animCurve = animCurve,
        repeatWrapperRot = repeatWrapperRot
      })
      end
    end
    wrapAnimNodeI = wrapAnimNodeI + 1
  end
  self.wrapAnimNodes = wrapAnimNodes
  local wrapNodes = {}
  local wrapNodeI = 0
  while true do
    local wrapNodeKey = string.format("vehicle.wrapper.wrapNode(%d)", wrapNodeI)
    if not hasXMLProperty(xmlFile, wrapNodeKey) then
      break
    end
    local nodeId = Utils.indexToObject(self.components, getXMLString(xmlFile, wrapNodeKey .. "#index"))
    local wrapVisibility = Utils.getNoNil(getXMLBool(xmlFile, wrapNodeKey .. "#wrapVisibility"), false)
    local emptyVisibility = Utils.getNoNil(getXMLBool(xmlFile, wrapNodeKey .. "#emptyVisibility"), false)
    if nodeId ~= nil and (wrapVisibility or emptyVisibility) then
      local maxWrapperRot = getXMLFloat(xmlFile, wrapNodeKey .. "#maxWrapperRot")
      if maxWrapperRot == nil then
        maxWrapperRot = math.huge
      else
        maxWrapperRot = math.rad(maxWrapperRot)
      end
      table.insert(wrapNodes, {
        nodeId = nodeId,
        wrapVisibility = wrapVisibility,
        emptyVisibility = emptyVisibility,
        maxWrapperRot = maxWrapperRot
      })
    end
    wrapNodeI = wrapNodeI + 1
  end
  self.wrapNodes = wrapNodes
  self:updateWrapNodes(false, true, 0)
  self.baleGrabber = {}
  self.baleGrabber.grabNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.baleGrabber#index"))
  self.baleGrabber.nearestDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.baleGrabber#nearestDistance"), 3)
  if self.isClient then
    self.wrapperSound = {}
    Utils.loadSample(xmlFile, self.wrapperSound, "vehicle.wrapperSound", nil, self.baseDirectory)
    self.wrapperStartSound = {}
    Utils.loadSample(xmlFile, self.wrapperStartSound, "vehicle.wrapperStartSound", nil, self.baseDirectory)
    self.wrapperStopSound = {}
    Utils.loadSample(xmlFile, self.wrapperStopSound, "vehicle.wrapperStopSound", nil, self.baseDirectory)
  end
  self.baleToLoad = nil
  self.baleToMount = nil
  self.baleWrapperState = BaleWrapper.STATE_NONE
  self.grabberIsMoving = false
  self.hasBaleWrapper = true
end
function BaleWrapper:delete()
  if self.isClient then
    Utils.deleteSample(self.wrapperSound)
    Utils.deleteSample(self.wrapperStartSound)
    Utils.deleteSample(self.wrapperStopSound)
  end
end
function BaleWrapper:readStream(streamId, connection)
  if connection:getIsServer() then
    local wrapperState = streamReadUIntN(streamId, BaleWrapper.STATE_NUM_BITS)
    if wrapperState >= BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER then
      local baleServerId
      if wrapperState ~= BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM then
        baleServerId = streamReadInt32(streamId)
      end
      if wrapperState == BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER then
        self:doStateChange(BaleWrapper.CHANGE_GRAB_BALE, baleServerId)
        AnimatedVehicle.updateAnimations(self, 99999999)
      elseif wrapperState == BaleWrapper.STATE_MOVING_GRABBER_TO_WORK then
        self.baleGrabber.currentBale = baleServerId
        self:doStateChange(BaleWrapper.CHANGE_DROP_BALE_AT_GRABBER)
        AnimatedVehicle.updateAnimations(self, 99999999)
      elseif wrapperState ~= BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM then
        local attachNode = self.wrapper.baleNode
        self.baleToMount = {
          serverId = baleServerId,
          linkNode = attachNode,
          trans = {
            0,
            0,
            0
          },
          rot = {
            0,
            0,
            0
          }
        }
        self:updateWrapNodes(true, false, 0)
        self.wrapper.currentBale = baleServerId
        if wrapperState == BaleWrapper.STATE_WRAPPER_WRAPPING_BALE then
          local wrapperTime = streamReadFloat32(streamId)
          self.wrapper.currentTime = wrapperTime
          self:updateWrappingState(self.wrapper.currentTime / self.wrapper.animTime, true)
        else
          self.wrapper.currentTime = self.wrapper.animTime
          self:updateWrappingState(1, true)
          self:doStateChange(BaleWrapper.CHANGE_WRAPPING_BALE_FINSIHED)
          AnimatedVehicle.updateAnimations(self, 99999999)
          if wrapperState >= BaleWrapper.STATE_WRAPPER_DROPPING_BALE then
            self:doStateChange(BaleWrapper.CHANGE_WRAPPER_START_DROP_BALE)
            AnimatedVehicle.updateAnimations(self, 99999999)
          end
        end
      else
        self.baleWrapperState = BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM
      end
    end
  end
end
function BaleWrapper:writeStream(streamId, connection)
  if not connection:getIsServer() then
    local wrapperState = self.baleWrapperState
    streamWriteUIntN(streamId, wrapperState, BaleWrapper.STATE_NUM_BITS)
    if wrapperState >= BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER and wrapperState ~= BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM then
      if wrapperState == BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER then
        streamWriteInt32(streamId, self.baleGrabber.currentBale)
      else
        streamWriteInt32(streamId, self.wrapper.currentBale)
      end
    end
    if wrapperState == BaleWrapper.STATE_WRAPPER_WRAPPING_BALE then
      streamWriteFloat32(streamId, self.wrapper.currentTime)
    end
  end
end
function BaleWrapper:mouseEvent(posX, posY, isDown, isUp, button)
end
function BaleWrapper:keyEvent(unicode, sym, modifier, isDown)
end
function BaleWrapper:update(dt)
  if self.firstTimeRun then
    if self.baleToLoad ~= nil then
      local v = self.baleToLoad
      self.baleToLoad = nil
      local baleObject = Bale:new(self.isServer, self.isClient)
      local x, y, z = unpack(v.translation)
      local rx, ry, rz = unpack(v.rotation)
      baleObject:load(v.filename, x, y, z, rx, ry, rz, v.fillLevel)
      if baleObject.nodeId ~= nil and baleObject.nodeId ~= 0 then
        baleObject.baleValueScale = v.baleValueScale
        local wrapperState = math.min(v.wrapperTime / self.wrapper.animTime, 1)
        baleObject:setWrappingState(wrapperState)
        baleObject:mount(self, v.parentNode, x, y, z, rx, ry, rz)
        baleObject:register()
        self:doStateChange(BaleWrapper.CHANGE_WRAPPING_START)
        self.wrapper.currentBale = networkGetObjectId(baleObject)
        self.wrapper.currentTime = v.wrapperTime
        self:updateWrappingState(self.wrapper.currentTime / self.wrapper.animTime)
      end
    end
    if self.baleToMount ~= nil then
      local bale = networkGetObject(self.baleToMount.serverId)
      if bale ~= nil then
        local x, y, z = unpack(self.baleToMount.trans)
        local rx, ry, rz = unpack(self.baleToMount.rot)
        bale:mount(self, self.baleToMount.linkNode, x, y, z, rx, ry, rz)
        self.baleToMount = nil
      end
    end
  end
  if self:getIsActive() then
    if self:getIsActiveForInput() and self.baleWrapperState == BaleWrapper.STATE_WRAPPER_FINSIHED and InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
      g_client:getServerConnection():sendEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_BUTTON_EMPTY))
    end
    if self.baleWrapperState == BaleWrapper.STATE_WRAPPER_WRAPPING_BALE then
      if self.isClient and self:getIsActiveForSound() and not self.wrapperSound.isPlaying and (self.wrapperStartSound.sample == nil or not isSamplePlaying(self.wrapperStartSound.sample) or getSamplePlayOffset(self.wrapperStartSound.sample) >= self.wrapperStartSound.duration - 1.8 * dt) then
        Utils.playSample(self.wrapperSound, 0, 0)
      end
      self.wrapper.currentTime = self.wrapper.currentTime + dt
      self:updateWrappingState(self.wrapper.currentTime / self.wrapper.animTime)
    end
  end
end
function BaleWrapper:updateTick(dt)
  if self:getIsActive() then
    self.showInvalidBaleWarning = false
    if self:allowsGrabbingBale() and self.baleGrabber.grabNode ~= nil and self.baleGrabber.currentBale == nil then
      local nearestBale, nearestBaleType = BaleWrapper.getBaleInRange(self, self.baleGrabber.grabNode)
      if nearestBale ~= nil then
        if nearestBaleType == nil then
          self.showInvalidBaleWarning = true
        elseif self.isServer then
          if nearestBale.i3dFilename ~= nearestBaleType.wrapperBaleFilename then
            local fillLevel = nearestBale.fillLevel
            local baleValueScale = nearestBale.baleValueScale
            nearestBale:delete()
            nearestBale = Bale:new(self.isServer, self.isClient)
            nearestBale:load(nearestBaleType.wrapperBaleFilename, 0, 0, 0, 0, 0, 0, fillLevel)
            nearestBale.baleValueScale = baleValueScale
            nearestBale:register()
          end
          g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_GRAB_BALE, networkGetObjectId(nearestBale)), true, nil, self)
        end
      end
    end
    if self.isServer and self.baleWrapperState ~= BaleWrapper.STATE_NONE then
      if self.baleWrapperState == BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER then
        if not self:getIsAnimationPlaying("moveBaleToWrapper") then
          g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_DROP_BALE_AT_GRABBER), true, nil, self)
        end
      elseif self.baleWrapperState == BaleWrapper.STATE_MOVING_GRABBER_TO_WORK then
        if not self:getIsAnimationPlaying("moveBaleToWrapper") then
          g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPING_START), true, nil, self)
        end
      elseif self.baleWrapperState == BaleWrapper.STATE_WRAPPER_DROPPING_BALE then
        if not self:getIsAnimationPlaying("dropBaleFromWrapper") then
          g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPER_BALE_DROPPED), true, nil, self)
        end
      elseif self.baleWrapperState == BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM and not self:getIsAnimationPlaying("resetWrapperAfterBaleDrop") then
        g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPER_PLATFORM_RESET), true, nil, self)
      end
    end
  end
end
function BaleWrapper:draw()
  if self.isClient then
    if self:getIsActiveForInput(true) and self.baleWrapperState == BaleWrapper.STATE_WRAPPER_FINSIHED then
      g_currentMission:addHelpButtonText(g_i18n:getText("BaleWrapper_unload"), InputBinding.IMPLEMENT_EXTRA3)
    end
    if self.showInvalidBaleWarning then
      g_currentMission:addWarning(g_i18n:getText("InvalidBaleWarning"))
    end
  end
end
function BaleWrapper:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
  if not resetVehicles then
    local filename = getXMLString(xmlFile, key .. "#baleFileName")
    if filename ~= nil then
      filename = Utils.convertFromNetworkFilename(filename)
      local wrapperTime = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#wrapperTime"), 0)
      local baleValueScale = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#baleValueScale"), 1)
      local fillLevel = getXMLFloat(xmlFile, key .. "#fillLevel")
      local translation = {
        0,
        0,
        0
      }
      local rotation = {
        0,
        0,
        0
      }
      local parentNode = self.wrapper.baleNode
      self.baleToLoad = {
        parentNode = parentNode,
        filename = filename,
        translation = translation,
        rotation = rotation,
        fillLevel = fillLevel,
        wrapperTime = wrapperTime,
        baleValueScale = baleValueScale
      }
    end
  end
  return BaseMission.VEHICLE_LOAD_OK
end
function BaleWrapper:getSaveAttributesAndNodes(nodeIdent)
  local attributes = ""
  local baleServerId = self.baleGrabber.currentBale
  if baleServerId == nil then
    baleServerId = self.wrapper.currentBale
  end
  if baleServerId ~= nil then
    local bale = networkGetObject(baleServerId)
    if bale ~= nil then
      local fillLevel = bale:getFillLevel()
      local baleValueScale = bale.baleValueScale
      attributes = "baleFileName=\"" .. Utils.encodeToHTML(Utils.convertToNetworkFilename(bale.i3dFilename)) .. "\" fillLevel=\"" .. fillLevel .. "\" wrapperTime=\"" .. tostring(self.wrapper.currentTime) .. "\" baleValueScale=\"" .. baleValueScale .. "\""
    end
  end
  return attributes, nil
end
function BaleWrapper:onDetach()
  if self.deactivateOnDetach then
    BaleWrapper.onDeactivate(self)
  else
    BaleWrapper.onDeactivateSounds(self)
  end
end
function BaleWrapper:onLeave()
  if self.deactivateOnLeave then
    BaleWrapper.onDeactivate(self)
  else
    BaleWrapper.onDeactivateSounds(self)
  end
end
function BaleWrapper:onDeactivate()
  BaleWrapper.onDeactivateSounds(self)
  self.showInvalidBaleWarning = false
end
function BaleWrapper:onDeactivateSounds()
  if self.isClient then
    Utils.stopSample(self.wrapperStartSound)
    Utils.stopSample(self.wrapperStopSound)
    Utils.stopSample(self.wrapperSound)
  end
end
function BaleWrapper:allowsGrabbingBale()
  local foldAnimTime = self.foldAnimTime
  if foldAnimTime ~= nil and (foldAnimTime > self.wrapper.foldMaxLimit or foldAnimTime < self.wrapper.foldMinLimit) then
    return false
  end
  return self.baleWrapperState == BaleWrapper.STATE_NONE
end
function BaleWrapper:updateWrapNodes(isWrapping, isEmpty, wrapperRot)
  for _, wrapNode in pairs(self.wrapNodes) do
    setVisibility(wrapNode.nodeId, (isWrapping and wrapNode.wrapVisibility or isEmpty and wrapNode.emptyVisibility) and wrapperRot < wrapNode.maxWrapperRot)
  end
  if isWrapping then
    local wrapperRotRepeat = wrapperRot % math.pi
    if wrapperRotRepeat < 0 then
      wrapperRotRepeat = wrapperRotRepeat + math.pi
    end
    for _, wrapAnimNode in pairs(self.wrapAnimNodes) do
      local rot = wrapperRot
      if wrapAnimNode.repeatWrapperRot then
        rot = wrapperRotRepeat
      end
      local v = wrapAnimNode.animCurve:get(rot)
      setTranslation(wrapAnimNode.nodeId, v[1], v[2], v[3])
      setRotation(wrapAnimNode.nodeId, v[4], v[5], v[6])
      setScale(wrapAnimNode.nodeId, v[7], v[8], v[9])
    end
  end
end
function BaleWrapper:updateWrappingState(t, noEventSend)
  t = math.min(t, 1)
  local wrapperRot = 0
  if self.wrapper.animCurve ~= nil then
    local v = self.wrapper.animCurve:get(t)
    setRotation(self.wrapper.baleNode, v[1] % (math.pi * 2), v[2] % (math.pi * 2), v[3] % (math.pi * 2))
    setRotation(self.wrapper.wrapperNode, v[4] % (math.pi * 2), v[5] % (math.pi * 2), v[6] % (math.pi * 2))
    wrapperRot = v[5]
    if self.wrapper.currentBale ~= nil and self.isServer then
      local bale = networkGetObject(self.wrapper.currentBale)
      if bale ~= nil then
        bale:setWrappingState(t)
      end
    end
  end
  self:updateWrapNodes(t > 0, false, wrapperRot)
  if t == 1 and self.isServer and self.baleWrapperState == BaleWrapper.STATE_WRAPPER_WRAPPING_BALE and not noEventSend then
    g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPING_BALE_FINSIHED), true, nil, self)
  end
end
function BaleWrapper:moveBaleToWrapper(bale)
  self:doStateChange(BaleWrapper.CHANGE_GRAB_BALE, networkGetObjectId(bale))
end
function BaleWrapper:doStateChange(id, nearestBaleServerId)
  if id == BaleWrapper.CHANGE_GRAB_BALE then
    local bale = networkGetObject(nearestBaleServerId)
    self.baleGrabber.currentBale = nearestBaleServerId
    if bale ~= nil then
      bale:mount(self, self.baleGrabber.grabNode, 0, 0, 0, 0, 0, 0)
      self.baleToMount = nil
    else
      self.baleToMount = {
        serverId = nearestBaleServerId,
        linkNode = self.baleGrabber.grabNode,
        trans = {
          0,
          0,
          0
        },
        rot = {
          0,
          0,
          0
        }
      }
    end
    self.baleWrapperState = BaleWrapper.STATE_MOVING_BALE_TO_WRAPPER
    self:playAnimation("moveBaleToWrapper", 1, nil, true)
  elseif id == BaleWrapper.CHANGE_DROP_BALE_AT_GRABBER then
    local attachNode = self.wrapper.baleNode
    local bale = networkGetObject(self.baleGrabber.currentBale)
    if bale ~= nil then
      bale:mount(self, attachNode, 0, 0, 0, 0, 0, 0)
      self.baleToMount = nil
    else
      self.baleToMount = {
        serverId = self.baleGrabber.currentBale,
        linkNode = attachNode,
        trans = {
          0,
          0,
          0
        },
        rot = {
          0,
          0,
          0
        }
      }
    end
    self:updateWrapNodes(true, false, 0)
    self.wrapper.currentBale = self.baleGrabber.currentBale
    self.baleGrabber.currentBale = nil
    self:playAnimation("moveBaleToWrapper", -1, nil, true)
    self.baleWrapperState = BaleWrapper.STATE_MOVING_GRABBER_TO_WORK
  elseif id == BaleWrapper.CHANGE_WRAPPING_START then
    self.baleWrapperState = BaleWrapper.STATE_WRAPPER_WRAPPING_BALE
    if self.isClient and self:getIsActiveForSound() then
      Utils.playSample(self.wrapperStartSound, 1, 0)
    end
  elseif id == BaleWrapper.CHANGE_WRAPPING_BALE_FINSIHED then
    if self.isClient then
      Utils.stopSample(self.wrapperSound)
      if self.isClient and self:getIsActiveForSound() then
        Utils.playSample(self.wrapperStopSound, 1, 0)
      end
    end
    self:updateWrappingState(1, true)
    self.baleWrapperState = BaleWrapper.STATE_WRAPPER_FINSIHED
  elseif id == BaleWrapper.CHANGE_WRAPPER_START_DROP_BALE then
    self:updateWrapNodes(false, false, 0)
    self:playAnimation("dropBaleFromWrapper", 1, nil, true)
    self.baleWrapperState = BaleWrapper.STATE_WRAPPER_DROPPING_BALE
  elseif id == BaleWrapper.CHANGE_WRAPPER_BALE_DROPPED then
    local bale = networkGetObject(self.wrapper.currentBale)
    if bale ~= nil then
      bale:unmount()
    end
    self.wrapper.currentBale = nil
    self.wrapper.currentTime = 0
    self:playAnimation("resetWrapperAfterBaleDrop", 1, nil, true)
    self.baleWrapperState = BaleWrapper.STATE_WRAPPER_RESETTING_PLATFORM
  elseif id == BaleWrapper.CHANGE_WRAPPER_PLATFORM_RESET then
    self:updateWrappingState(0)
    self:updateWrapNodes(false, true, 0)
    self.baleWrapperState = BaleWrapper.STATE_NONE
  elseif id == BaleWrapper.CHANGE_BUTTON_EMPTY then
    assert(self.isServer)
    if self.baleWrapperState == BaleWrapper.STATE_WRAPPER_FINSIHED then
      g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPER_START_DROP_BALE), true, nil, self)
    end
  end
end
function BaleWrapper:setIsUnloadingBale(isUnloadingBale, noEventSend)
  if self.baleUnloadAnimationName ~= nil then
    if isUnloadingBale then
      if self.BaleWrapperUnloadingState ~= BaleWrapper.UNLOADING_OPENING then
        BaleWrapperSetIsUnloadingBaleEvent.sendEvent(self, isUnloadingBale, noEventSend)
        self.BaleWrapperUnloadingState = BaleWrapper.UNLOADING_OPENING
        self:playAnimation(self.baleUnloadAnimationName, self.baleUnloadAnimationSpeed, nil, true)
      end
    elseif self.BaleWrapperUnloadingState ~= BaleWrapper.UNLOADING_CLOSING then
      BaleWrapperSetIsUnloadingBaleEvent.sendEvent(self, isUnloadingBale, noEventSend)
      self.BaleWrapperUnloadingState = BaleWrapper.UNLOADING_CLOSING
      self:playAnimation(self.baleCloseAnimationName, self.baleCloseAnimationSpeed, nil, true)
    end
  end
end
function BaleWrapper:getBaleInRange(refNode)
  local px, py, pz = getWorldTranslation(refNode)
  local nearestDistance = 2
  local nearestBale, nearestBaleType
  for index, item in pairs(g_currentMission.itemsToSave) do
    local bale = item.item
    if bale:isa(Bale) then
      local vx, vy, vz = getWorldTranslation(bale.nodeId)
      local distance = Utils.vector3Length(px - vx, py - vy, pz - vz)
      if nearestDistance > distance then
        local foundBaleType
        if not bale.supportsWrapping or bale.wrappingState < 0.99 then
          local baleTypes = self.allowedBaleTypes[bale:getFillType()]
          if baleTypes ~= nil then
            for _, baleType in pairs(baleTypes) do
              if bale.baleDiameter ~= nil and bale.baleWidth ~= nil and bale.baleDiameter >= baleType.minBaleDiameter and bale.baleDiameter <= baleType.maxBaleDiameter and bale.baleWidth >= baleType.minBaleWidth and bale.baleWidth <= baleType.maxBaleWidth then
                foundBaleType = baleType
                break
              end
            end
          end
        end
        if foundBaleType ~= nil or nearestBaleType == nil then
          if foundBaleType ~= nil then
            nearestDistance = distance
          end
          nearestBale = bale
          nearestBaleType = foundBaleType
        end
      end
    end
  end
  return nearestBale, nearestBaleType
end
function BaleWrapper:getIsFoldAllowed(superFunc)
  if self.baleWrapperState ~= BaleWrapper.STATE_NONE then
    return false
  end
  if superFunc ~= nil then
    return superFunc(self)
  end
  return true
end
--local modFilename, isMod, isDlc, dlcsDirectoryIndex = Utils.removeModDirectory(g_currentModDirectory)
--if not isDlc then
--  BaleWrapper = nil
--end
